
const {
    get_local_wd,
    get_utc_wd,
    _wd_to_str,
} = require("./week")



function _prepend(n,width,padding) {
    let s = n.toString()
    padding = padding.toString()
    let lngth = s.length
    s = padding.repeat(width-lngth) + s
    return(s.substring(0,width))
}


function _moffset_to_z(moffset) {
    let min = moffset % 60
    let h = (moffset - min) / 60
    min = _prepend(min,2,0)
    h = _prepend(h,2,0)
    let s = (moffset>0)?'+':'-'
    return(s+h+min)
}


function _z_to_moffset(z) {
    let s = z[0]
    let h = parseInt(z.slice(1,3))
    let m = parseInt(z.slice(3,5))
    m = h * 60 + m
    m = (s==='+')?m:-m
    return(m)
}

function _get_z(dt) {
    let moffset = - dt.getTimezoneOffset()
    let z = _moffset_to_z(moffset)
    return(z)
}


function _dt_to_dict(dt,zone) {
    const moffset = _z_to_moffset(zone)
    const soffset = moffset * 60;
    const msoffset = soffset * 1000;
    ////////////////////////////////////////
    const mts =   dt.getTime();
    const ts =        mts /1000;
    //////////////////////////////////
    const dummy_mts = mts + msoffset
    const dummy_dt = (new Date(dummy_mts));
    const y =     dummy_dt.getUTCFullYear();
    const m =     dummy_dt.getUTCMonth() + 1;
    const d =     dummy_dt.getUTCDate();
    const h =     dummy_dt.getUTCHours();
    const min =   dummy_dt.getUTCMinutes();
    const s =     dummy_dt.getUTCSeconds();
    const ms =    dummy_dt.getUTCMilliseconds();
    ////
    const zone_desc = dt.toString().split(/( \()|(\))/)[3];
    ////
    const utc_wd = get_utc_wd(dummy_dt)
    const lcl_wd = get_local_wd(dummy_dt);
    ////
    let dict = {
        y,m,d,
        h,min,s,ms,
        ts,mts,
        zone,zone_desc,
	soffset,msoffset,
        utc_wd,lcl_wd,
    }
    return(dict)
}


function _year_to_str(y) {return(_prepend(y,4,0))}

function _month_to_str(m) {return(_prepend(m,2,0))}

function _day_to_str(d) {return(_prepend(d,2,0))}


function _ymd_to_str(y,m,d) {
    y = _year_to_str(y)
    m = _month_to_str(m)
    d = _day_to_str(d)
    return(y+'-'+m+'-'+d)
}


function _hour_to_str(h) {return(_prepend(h,2,0))}

function _min_to_str(min) {return(_prepend(min,2,0))}

function _sec_to_str(sec) {return(_prepend(sec,2,0))}

function _hms_to_str(h,min,sec) {
    let hstr = _hour_to_str(h)
    let mstr = _min_to_str(min)
    let sstr = _sec_to_str(sec)
    return(hstr+':'+mstr+':'+sstr)
}

function _msec_to_str(msec) {return(_prepend(msec,3,0))}

function _dict_to_str(dict) {
    //fmt='YYYY-MM-DD HH:mm:ss Z'
    let ymd_str = _ymd_to_str(dict.y,dict.m,dict.d)
    let hms_str = _hms_to_str(dict.h,dict.min,dict.s)
    z = dict.zone
    return(ymd_str+' '+hms_str+'.'+_msec_to_str(dict.ms)+' '+z)
}

function _dt_to_str(dt,z) {
    //fmt='YYYY-MM-DD HH:mm:ss Z'
    let dict = _dt_to_dict(dt,z)
    return(_dict_to_str(dict))
}

function _dt_to_sdict(dt,z) {
    let dict = _dt_to_dict(dt,z);
    dict.y = _year_to_str(dict.y);
    dict.m = _month_to_str(dict.m);
    dict.d = _day_to_str(dict.d);
    dict.h = _hour_to_str(dict.h);
    dict.min = _min_to_str(dict.min);
    dict.s = _sec_to_str(dict.s);
    dict.ms = _msec_to_str(dict.ms);
    ////
    dict.utc_wd = _wd_to_str(dict.utc_wd) 
    dict.lcl_wd = _wd_to_str(dict.lcl_wd)    
    return(dict)
}



class Auto {
    #z
    constructor(z) {
        if(z===undefined) {
            let dt = new Date();
            this.#z = _get_z(dt);
        } else {
            this.#z = z
        }
    }
    get zone() {return(this.#z)}
    set zone(z) {this.#z =z}
    dict() {return(_dt_to_dict(new Date(),this.#z))}
    str() {return(_dt_to_str(new Date(),this.#z))}
    sdict() { return(_dt_to_sdict(new Date(),this.#z))}
    get [Symbol.toStringTag]() {
        return(_dt_to_str(new Date(),this.#z))
    }
    clone() {
        let mts = this.dict().mts;
        let manual = new Manual(mts);
        manual.zone = this.#z;
        return(manual)
    }
}

const creat_auto = (o) => new Auto(o);


const TOTAL_DAYS_DICT_EXCEPT_FEB = {
    1:31,3:31,4:30,5:31,
    6:30,7:31,8:31,9:30,
    10:31,11:30,12:31
}

const _fst_yd = (that)=> {
        let y = that.sdict().y;
        let d = new Manual(y+"-01-01 00:00:00 "+that.zone);
        return(d)
}

const _lst_yd = (that)=> {
        let y = that.sdict().y;
        let d = new Manual(y+"-12-31 00:00:00 "+that.zone);
        return(d)
}

const _fst_md = (that)=> {
        let od = that.sdict();
        let y = od.y;
        let m = od.m;
        let d = new Manual(y+`-${m}-01 00:00:00 `+that.zone);
        return(d)
}

const _lst_md = (that)=> {
      let od = that.sdict();
      let y = od.y;
      let m = od.m;
      let dofm = that.lst_day_str_of_this_month_;
      let d = new Manual(y+`-${m}-${dofm} 00:00:00 `+that.zone);
      return(d)
}


const FST_QM = {1:1,2:4,3:7,4:10}
const LST_QM = {1:3,2:6,3:9,4:12}


class Manual {
    #tm
    #z
    constructor(...params) {
        let dt = new Date(...params);
        this.#tm = dt.getTime();
        this.#z = _get_z(dt)
    }
    get zone() {return(this.#z)}
    set zone(z) {this.#z =z}
    dict() {return(_dt_to_dict(new Date(this.#tm),this.#z))}
    str() {return(_dt_to_str(new Date(this.#tm),this.#z))}
    sdict() { return(_dt_to_sdict(new Date(this.#tm),this.#z))}
    date()  {return(new Date(this.#tm))}
    get [Symbol.toStringTag]() {
        return(_dt_to_str(new Date(this.#tm),this.#z))
    }
    next_ms(n=1) {
        this.#tm = this.#tm+1*n;
        return(_dt_to_str(new Date(this.#tm),this.#z))
    }
    prev_ms(n=1) {
        this.#tm = this.#tm-1*n;
        return(_dt_to_str(new Date(this.#tm),this.#z))
    }
    next_s(n=1) {
        this.#tm = this.#tm+1000*n;
        return(_dt_to_str(new Date(this.#tm),this.#z))
    }
    prev_s(n=1) {
        this.#tm = this.#tm-1000*n;
        return(_dt_to_str(new Date(this.#tm),this.#z))
    }
    next_min(n=1) {
        this.#tm = this.#tm+1000*60*n;
        return(_dt_to_str(new Date(this.#tm),this.#z))
    }
    prev_min(n=1) {
        this.#tm = this.#tm-1000*60*n;
        return(_dt_to_str(new Date(this.#tm),this.#z))
    }
    next_h(n=1) {
        this.#tm = this.#tm+1000*60*60*n;
        return(_dt_to_str(new Date(this.#tm),this.#z))
    }
    prev_h(n=1) {
        this.#tm = this.#tm-1000*60*60*n;
        return(_dt_to_str(new Date(this.#tm),this.#z))
    }
    next_d(n=1) {
        this.#tm = this.#tm+1000*60*60*24*n;
        return(_dt_to_str(new Date(this.#tm),this.#z))
    }
    prev_d(n=1) {
        this.#tm = this.#tm-1000*60*60*24*n;
        return(_dt_to_str(new Date(this.#tm),this.#z))
    }
    next_w(n=1) {
        this.#tm = this.#tm+1000*60*60*24*7*n;
        return(_dt_to_str(new Date(this.#tm),this.#z))
    }
    prev_w(n=1) {
        this.#tm = this.#tm-1000*60*60*24*7*n;
        return(_dt_to_str(new Date(this.#tm),this.#z))
    }
    next_m(n=1) {
        ////
        let r = n%12;
        let q = (n-r) / 12;
        ////
        let d = this.dict();
        ////
        d.y = d.y + q;
        if(r===0) {
        } else {
            let m = d.m + r;
            if(m<=12) {
                d.m=m
            } else {
                d.m = m -12
                d.y = d.y + 1
            }
        }
        ////
        let s = _dict_to_str(d);
        let dt = new Date(s);
        this.#tm = dt.getTime();
        return(s)
    }
    prev_m(n=1) {
        ////
        let r = n%12;
        let q = (n-r) / 12;
        ////
        let d = this.dict();
        if(r===0) {
        } else {
            let m = d.m - r
            if(m>=1) {
                d.m = m
            } else {
                d.y = d.y -1;
                m = m +12 
                d.m = m
            }
        }
        d.y = d.y - q;
        ////
        let s = _dict_to_str(d);
        let dt = new Date(s);
        this.#tm = dt.getTime();
        return(s)
    }
    next_y(n=1) {
        let d = this.dict();
        d.y = d.y + n;
        let s = _dict_to_str(d);
        let dt = new Date(s);
        this.#tm = dt.getTime();
        return(s)
    }
    prev_y(n=1) {
        let d = this.dict();
        d.y = d.y - n;
        let s = _dict_to_str(d);
        let dt = new Date(s);
        this.#tm = dt.getTime();
        return(s)
    }
    ////
    clone() {
        let mts = this.dict().mts;
        let manual = new Manual(mts);
        manual.zone = this.#z;
        return(manual)
    }
    ////
    * nms_gen$$  (n=1) {while(true) {yield(this.next_ms(n))}}
    * pms_gen$$  (n=1) {while(true) {yield(this.prev_ms(n))}}
    * ns_gen$$  (n=1) {while(true) {yield(this.next_s(n))}}
    * ps_gen$$  (n=1) {while(true) {yield(this.prev_s(n))}}
    * nmin_gen$$  (n=1) {while(true) {yield(this.next_min(n))}}
    * pmin_gen$$  (n=1) {while(true) {yield(this.prev_min(n))}}
    * nh_gen$$  (n=1) {while(true) {yield(this.next_h(n))}}
    * ph_gen$$  (n=1) {while(true) {yield(this.prev_h(n))}}
    * nd_gen$$  (n=1) {while(true) {yield(this.next_d(n))}}
    * pd_gen$$  (n=1) {while(true) {yield(this.prev_d(n))}}
    * nm_gen$$  (n=1) {while(true) {yield(this.next_m(n))}}
    * pm_gen$$  (n=1) {while(true) {yield(this.prev_m(n))}}
    * ny_gen$$  (n=1) {while(true) {yield(this.next_h(n))}}
    * py_gen$$  (n=1) {while(true) {yield(this.prev_h(n))}}
    [Symbol.iterator]() {return(this.nms_gen$$())}
    /////
    is_leap_year() {
        let d = this.dict();
	let y = d.y;
        return((y%4===0 && y%100!==0 )|| (y%400===0))
    }
    get fst_yd_() {return(_fst_yd(this))}
    get lst_yd_() {return(_lst_yd(this))}
    get total_days_of_this_year_() { return(this.is_leap_year()?366:365)}
    ////
    get q_()  {
	let d = this.dict();
	let m = d.m;
	m = m - 1;
	return((m - m % 3)/3 + 1)
    }
    get fst_qm_()  {
	let d = this.dict();
	let m = d.m;
	return(FST_QM[m])
    }
    get lst_qm_()  {
        let d = this.dict();
        let m = d.m;
        return(LST_QM[m])
    }
    get fst_qd_() {
	let d = this.dict();  let y = d.y;
        let qm = this.fst_qm_;
	let that = new Manual(y+`-${qm}-01 00:00:00 `+that.zone);
	return(that.fst_md_)
    }
    get lst_qd_() {
        let d = this.dict();  let y = d.y;
        let qm = this.lst_qm_;
        let that = new Manual(y+`-${qm}-01 00:00:00 `+that.zone);
        return(that.lst_md_)
    }
    ////
    get lst_day_of_this_month_() {
        let sd = this.sdict();
	let d  = this.dict();
	let days;
        if(d.m===2) {
	    days = (this.is_leap_year())?29:28;  	
	} else {
	    days = TOTAL_DAYS_DICT_EXCEPT_FEB[d.m]
	}
	return(days)
    }
    get lst_day_str_of_this_month_() {
        let days = this.lst_day_of_this_month_;
	if(days<10) {
            return("0"+String(days))
	} else {
            return(String(days))		
	}
    }
    get fst_md_() {return(_fst_md(this))}
    get lst_md_() {return(_lst_md(this))} 
    /////
    get fst_lcl_wd_()  {
        let d = this.dict();
	let wd = d.lcl_wd;  //1-7
        let diff = wd - 1;
	let mts = d.mts - (1000*3600*24)*diff;
	return(new Manual(mts))
    }
    get lst_lcl_wd_()  {
        let d = this.dict();
        let wd = d.lcl_wd;
        let diff = 7-wd;
        let mts = d.mts + (1000*3600*24)*diff;
        return(new Manual(mts))
    }
    get fst_utc_wd_()  {
        let d = this.dict();
        let wd = d.utc_wd;
        let diff = wd - 1;
        let mts = d.mts - (1000*3600*24)*diff;
        return(new Manual(mts))
    }       
    get lst_utc_wd_()  {
        let d = this.dict();
        let wd = d.utc_wd;
        let diff = 7-wd;
        let mts = d.mts + (1000*3600*24)*diff;
        return(new Manual(mts))
    }
}

const creat_manual = (o) => new Manual(o);



const ERROR_DICT = {
    running:"must-clear-before-refrequency"
}

const {
    set_interval_without_rslts,
    _psj,
} = require("nv-facutil-promise");


class Cron {
    #interval = null
    #frequency
    #cond_func
    #task_func
    #args
    #auto = new Auto()
    constructor(frequency,cond_func,task_func,...args) {
        this.#cond_func = cond_func;
        this.#task_func = task_func;
        this.#args = args
        this.#frequency = frequency;
    }
    get zone() {return(this.#auto.dict().zone)}
    set zone(z) {this.#auto.zone = z}
    get cond_func() {return(this.#cond_func)}
    set cond_func(f) {this.#cond_func =f }
    get task_func() {return(this.#task_func)}
    set task_func(f) {this.#task_func =f }
    get frequency() {return(this.#frequency)}
    set frequency(freq) {
        if(this.is_running()) {
            throw(ERROR_DICT.running)
        } else {
            this.#frequency = freq
        }
    }
    is_running(){return(this.#interval!==null)}
    clear() {
        this.#interval();
        this.#interval = null
    }
    launch() {
        let {newClearInterval} = set_interval_without_rslts(
             ()=> {
                 let d = this.#auto.dict();
                 let cond = this.#cond_func(d);
                 if(cond){this.#task_func(...this.#args)}
             },
             this.#frequency
        );
        this.#interval = newClearInterval;
    }
}

const now = ()=>(new Date()).getTime()


class StopWatch {
    #start_at = undefined;
    #end_at   = undefined;
    #psj = undefined;
    get start_at_()  {return(this.#start_at)}
    get end_at_()    {return(this.#end_at)}
    get passed_()    {
        if(this.#end_at !== undefined) {
            return(this.#end_at - this.#start_at)
        } else if(this.#start_at !== undefined) {
            return(now() - this.#start_at)
        } else {
        }
    }
    clear() {
        this.#start_at = undefined
        this.#end_at   = undefined
        this.#psj = undefined
    }
    start() {
        this.clear();
        this.#psj = _psj();
        this.#start_at = now();
        return(this.#psj[0])
    }
    stop() {
        this.#end_at = now();
        let d = {start_at:this.#start_at,end_at:this.#end_at,costed:this.passed_}
        this.#psj[1](d);
        return(d)
    }
    get [Symbol.toStringTag]() {
        let d = `start_at:${this.#start_at}  duration:${this.passed_}  end_at:${this.#end_at}`
        return(d)
    }
}


module.exports = {
    now,
    ////
    StopWatch,
    stop_watch:new StopWatch(),
    ////
    ERROR_DICT,
    Auto,      creat_auto,
    Manual,    creat_manual, 
    Cron,
    /////
    week:require("./week"),
    ////
    basic:require("./basic"),
}

